# Copyright 2024, Seiko Epson Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the “Software”), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

from dataclasses import dataclass
from typing import List, Optional, TypedDict, Union

from logger.constants import ConfigKeys, Constants
from logger.utils.env import get_int, get_str
from logger.utils.validator import Validator


class ConfigOverrideOption(TypedDict):
    """環境変数から設定値を読み込まず上書きするためのクラス"""

    measure_time_sec: Optional[int]


@dataclass
class Config:
    """設定値を保持するクラス"""

    logger_id: str
    output_path: str
    initial_wait: int
    measure_time_sec: int
    file_rotate_min: int
    baud_rate: int
    A342_physical: str
    A342_mode: str
    A342_rmspp_output_interval: int
    A352_sps: int


def get_config(option: ConfigOverrideOption) -> Union[Config, None]:
    """環境変数(設定ファイル)から設定値を読み込み設定オブジェクトを返す関数

    Args:
        option (ConfigOverrideOption): 環境変数からではなく読み込ませる設定値を渡すオブジェクト、現在は計測秒数のみ対応

    Returns:
        Union[Config, None]: 設定オブジェクト。バリデーションに失敗した場合はNoneが帰る
    """
    validator = Validator()
    results: List[bool] = []

    logger_id = get_str(ConfigKeys.LOGGER_ID)
    results.append(
        validator.match(logger_id, ConfigKeys.LOGGER_ID, Constants.LOGGER_ID_PATTERN)
    )

    output_path = get_str(ConfigKeys.OUTPUT_PATH)
    results.append(validator.exist_dir(output_path, ConfigKeys.OUTPUT_PATH))

    initial_wait = get_int(ConfigKeys.INITIAL_WAIT)
    results.append(
        validator.between(
            initial_wait,
            ConfigKeys.INITIAL_WAIT,
            Constants.INITIAL_WAIT_MIN,
            Constants.INITIAL_WAIT_MAX,
        )
    )

    measure_time_sec = (
        get_int(ConfigKeys.MEASURE_TIME_SEC)
        if option.get("measure_time_sec") is None
        else option.get("measure_time_sec")
    )
    results.append(
        validator.between(
            measure_time_sec,
            ConfigKeys.MEASURE_TIME_SEC,
            Constants.MEASURE_TIME_SEC_MIN,
            Constants.MEASURE_TIME_SEC_MAX,
        )
    )

    file_rotate_min = get_int(ConfigKeys.FILE_ROTATE_MIN)
    results.append(
        validator.between(
            file_rotate_min,
            ConfigKeys.FILE_ROTATE_MIN,
            Constants.FILE_ROTATE_MIN_MIN,
            Constants.FILE_ROTATE_MIN_MAX,
        )
    )

    baud_rate = get_int(ConfigKeys.BAUD_RATE)
    results.append(
        validator.one_of(baud_rate, ConfigKeys.BAUD_RATE, Constants.BAUD_RATES)
    )

    A342_physical = get_str(ConfigKeys.A342_PHYSICAL)
    results.append(
        validator.one_of(
            A342_physical, ConfigKeys.A342_PHYSICAL, Constants.A342_PHYSICALS
        )
    )

    A342_mode = get_str(ConfigKeys.A342_MODE)
    results.append(
        validator.one_of(A342_mode, ConfigKeys.A342_MODE, Constants.A342_MODES)
    )

    A342_rmspp_output_interval = get_int(ConfigKeys.A342_RMSPP_OUTPUT_INTERVAL)
    results.append(
        validator.between(
            A342_rmspp_output_interval,
            ConfigKeys.A342_RMSPP_OUTPUT_INTERVAL,
            Constants.A342_RMS_PP_OUTPUT_INTERVAL_MIN,
            Constants.A342_RMS_PP_OUTPUT_INTERVAL_MAX,
        )
    )

    A352_sps = get_int(ConfigKeys.A352_SPS)
    results.append(validator.one_of(A352_sps, ConfigKeys.A352_SPS, Constants.A352_SPSS))

    if not all(results):
        return None

    return Config(
        logger_id=logger_id,  # type: ignore
        output_path=output_path,  # type: ignore
        initial_wait=initial_wait,  # type: ignore
        measure_time_sec=measure_time_sec,  # type: ignore
        file_rotate_min=file_rotate_min,  # type: ignore
        baud_rate=baud_rate,  # type: ignore
        A342_physical=A342_physical,  # type: ignore
        A342_mode=A342_mode,  # type: ignore
        A342_rmspp_output_interval=A342_rmspp_output_interval,  # type: ignore
        A352_sps=A352_sps,  # type: ignore
    )
