# Copyright 2024, Seiko Epson Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the “Software”), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import math
from typing import List, Tuple, TypedDict

from logger.comm import Comm, WaitCommand
from logger.constants import Constants
from logger.measure.data import MeasureData
from logger.measure.job import ReaderArgs, WriterArgs
from logger.measure.sensor import Sensor
from logger.utils.convert import to_dec24, to_int8, to_int16, to_uint16


class A342ReaderArgs(ReaderArgs):
    def format_packet(self, index: int, packet: List[int]) -> MeasureData:
        """1レコード分のデータを受け取って計測データに変換するメソッド

        パケット構成: 13byte
        - 0: 0x80
        - 1-2: TEMP2
        - 3-5: XVELC
        - 6-8: YVELC
        - 9-11: ZVELC
        - 12: 0x0D

        パケット構成: 19byte
        - 0: 0x80
        - 1-2: FLAG
        - 3-4: TEMP1
        - 5-7: XVELC
        - 8-10: YVELC
        - 11-13: ZVELC
        - 14-15: COUNT
        - 16-17: CHECKSUM
        - 18: 0x0D

        Args:
            index (int): 何番目のデータか
            packet (List[int]): 変換対象のデータ配列

        Returns:
            MeasureData: 変換後のpythonデータクラス
        """
        if self.baud == Constants.BAUD_RATE_460800:
            # 460kbps通信時はフラグは温度2の下位バイトの上位6bit
            flag = packet[2] & 0b11111100
            temperature = to_int8(packet[1]) * -0.9707008 + 34.987
            x = to_dec24(packet[3], packet[4], packet[5])
            y = to_dec24(packet[6], packet[7], packet[8])
            z = to_dec24(packet[9], packet[10], packet[11])
            # 460kbps通信時はカウントは温度2の下位バイトの下位2bit
            count = packet[2] & 0b11
            return MeasureData(
                index=index,
                count=count,
                temperature=temperature,
                x=x,
                y=y,
                z=z,
                flag=flag,
            )
        else:
            flag = packet[2]
            temperature = to_int16(packet[3], packet[4]) * -0.0037918 + 34.987
            x = to_dec24(packet[5], packet[6], packet[7])
            y = to_dec24(packet[8], packet[9], packet[10])
            z = to_dec24(packet[11], packet[12], packet[13])
            count = to_uint16(packet[14], packet[15])
            return MeasureData(
                index=index,
                count=count,
                temperature=temperature,
                x=x,
                y=y,
                z=z,
                flag=flag,
            )

    def _gen_invalid_data(self, index: int, count: int) -> MeasureData:
        """データ抜けがあった際に不正な値であることを示すデータクラスオブジェクトを生成するメソッド

        Args:
            index (int): 不正な計測データのindex
            count (int): 不正な計測データのカウント値、complement_dataの中で前のデータから算出する

        Returns:
            MeasureData: 補完したindexとcountを持つ不正なことを表すデータクラスオブジェクト
        """
        invalid_value = to_dec24(0x7F, 0xFF, 0xFF)
        if self.baud == Constants.BAUD_RATE_460800:
            temperature = to_int8(0x7F) * -0.9707008 + 34.987
        else:
            temperature = to_int16(0x7F, 0xFF) * -0.0037918 + 34.987

        return MeasureData(
            index=index,
            count=count,
            temperature=temperature,
            x=invalid_value,
            y=invalid_value,
            z=invalid_value,
            flag=0,
        )


class BaudRateConfig(TypedDict):
    record_size: int
    tmp_command: int
    burst_ctl_command: Tuple[int, int]  # 順番に上位、下位
    count_max: int
    count_diff: int


class PhysicalConfig(TypedDict):
    sps: int
    command: int
    wait_sec: float


class ModeConfig(TypedDict):
    command: int


class A342(Sensor):
    """A342センサークラス"""

    BAUD_RATE_CONFIGS: dict[int, BaudRateConfig] = {
        Constants.BAUD_RATE_460800: {
            "record_size": 13,
            "tmp_command": 0b00,  # 温度2
            "burst_ctl_command": (
                0x47,
                0x00,
            ),  # flag: off, 温度: on(2), XYZ: on, カウントoff, チェックサムoff
            "count_max": 4,
            "count_diff": 1,
        },
        Constants.BAUD_RATE_921600: {
            "record_size": 19,
            "tmp_command": 0b10,  # 温度1
            "burst_ctl_command": (
                0xC7,
                0x03,
            ),  # flag: on, 温度: on(2), XYZ: on, カウントon, チェックサムon
            "count_max": 2**16,
            "count_diff": 2,
        },
    }
    PHYSICAL_CONFIGS: dict[str, PhysicalConfig] = {
        Constants.PHYSICAL_VELOCITY: {
            "sps": 3000,
            "command": 0b0000000,
            "wait_sec": 0.18,
        },
        Constants.PHYSICAL_DISPLACEMENT: {
            "sps": 300,
            "command": 0b1000000,
            "wait_sec": 1.74,
        },
    }
    MODE_CONFIGS: dict[str, ModeConfig] = {
        Constants.MODE_RAW: {"command": 0b000000},
        Constants.MODE_RMS: {"command": 0b010000},
        Constants.MODE_PP: {
            "command": 0b100000,
        },
    }

    def __init__(
        self,
        port: str,
        baud: int,
        product_id: str,
        logger_id: str,
        file_rotate_min: int,
        mode: str,
        physical: str,
        rms_pp_interval: int,
    ) -> None:
        super().__init__(
            model="A342",
            port=port,
            baud=baud,
            product_id=product_id,
            sps=A342.PHYSICAL_CONFIGS[physical]["sps"],
            logger_id=logger_id,
            file_rotate_min=file_rotate_min,
            physical=physical,
            mode=mode,
            rms_pp_interval=rms_pp_interval,
        )
        self.record_size = A342.BAUD_RATE_CONFIGS[baud]["record_size"]
        self.count_diff = A342.BAUD_RATE_CONFIGS[baud]["count_diff"]
        self.count_max = A342.BAUD_RATE_CONFIGS[baud]["count_max"]
        self.wait_command = WaitCommand(A342.PHYSICAL_CONFIGS[physical]["wait_sec"])

        # SIG_CONTROLのコマンドを合成する
        self.sig_ctl_command = (
            A342.PHYSICAL_CONFIGS[physical]["command"]
            | A342.MODE_CONFIGS[mode]["command"]
            | A342.BAUD_RATE_CONFIGS[baud]["tmp_command"]
        )
        self.burst_ctl_command = A342.BAUD_RATE_CONFIGS[baud]["burst_ctl_command"]
        self.smpl_ctl_command = self._get_smpl_ctl_command()

    def init(self) -> None:
        """与えられた計測設定に基づきセンサーの計測情報を設定するメソッド"""
        comm = Comm()
        comm.open(self.port, self.baud)
        comm.send(
            [
                [0, 0xFF, 0xFF, 0x0D],  # おまじない 3 回 （センサー復帰）
                [0, 0xFF, 0xFF, 0x0D],
                [0, 0xFF, 0xFF, 0x0D],
                # W: 出力タイプ
                [0, 0xFE, 0x01, 0x0D],  # Window 1
                [0, 0x80, self.sig_ctl_command, 0x0D],  # SIG_CTRL:
                [0, 0x81, 0x8E, 0x0D],  # SIG_CTRL: 15:TEMP 11-9:VELC
                WaitCommand(0.3),  # Wait > 118ms
                [
                    0,
                    0x84,
                    self.smpl_ctl_command[1],
                    0x0D,
                ],  # SMPL_CTRL: UPDATE_RATE_RMSPP
                [0, 0x85, self.smpl_ctl_command[0], 0x0D],  # SMPL_CTRL: DOUT_RATE_RMSPP
                # W: UART自動サンプリング
                [0, 0xFE, 0x01, 0x0D],  # Window 1
                [0, 0x88, 0x01, 0x0D],  # UART_CTRL: 1:AUTO_START=0 0:UART_AUTO=1
                [0, 0x8C, self.burst_ctl_command[1], 0x0D],  # BURST_CTRL: 下位
                [0, 0x8D, self.burst_ctl_command[0], 0x0D],  # BURST_CTRL: 上位
                # W: XYZ_ALARM -- 用途に応じ調整必要なため設定しない
                WaitCommand(0.1),  # Wait > 5ms
            ]
        )

    def to_reader_args(self) -> ReaderArgs:
        """読み込みプロセスに渡す設定情報に変換するメソッド

        Returns:
            ReaderArgs: 読み込みプロセスに渡す設定情報
        """

        assert self.rms_pp_interval
        # RMSとP-Pではデフォルトタイムアウトより長い期間readが帰らないため、計算する
        timeout: float = Comm.DEFAULT_TIMEOUT
        if self.mode == Constants.MODE_RMS or self.mode == Constants.MODE_PP:
            if self.physical == Constants.PHYSICAL_VELOCITY:
                timeout = self.rms_pp_interval / 10
            else:
                timeout = self.rms_pp_interval
        timeout = max(Comm.DEFAULT_TIMEOUT, timeout)

        count_start, count_diff = self._get_complement_data_params()
        record_per_sec = self._get_record_per_sec()
        # 一回の取得データ数が小さいとデータとびが発生するため、最大4096byteとして1秒間分のデータをまとめて読み込む
        read_length = min(record_per_sec * self.record_size, Comm.DEFAULT_READ_BYTES)

        return A342ReaderArgs(
            model=self.model,
            serial=self.serial,
            port=self.port,
            baud=self.baud,
            record_size=self.record_size,
            record_per_sec=record_per_sec,
            # 過渡応答期間があり、RMS, P-Pがタイムアウトする恐れがあるためスタート後にwaitを入れる
            start_command=[*Sensor.DEFAULT_START_COMMAND, self.wait_command],
            end_command=Sensor.DEFAULT_END_COMMAND,
            record_begin=Sensor.DEFAULT_RECORD_BEGIN,
            record_end=Sensor.DEFAULT_RECORD_END,
            read_length=read_length,
            timeout=timeout,
            count_diff=count_diff,
            count_max=self.count_max,
            count_start=count_start,
        )

    def to_writer_args(self) -> WriterArgs:
        """書き込みプロセスに渡す設定情報に変換するメソッド

        Returns:
            ReaderArgs: 書き込みプロセスに渡す設定情報
        """
        return WriterArgs(
            model=self.model,
            serial=self.serial,
            logger_id=self.logger_id,
            port=self.port,
            record_per_file=self._get_record_per_file(),
        )

    def _get_smpl_ctl_command(self) -> Tuple[int, int]:  # noqa: C901
        """RMSとP-Pの計算、出力間隔を取得するメソッド

        Returns:
            Tuple[int, int]: (出力レートのコマンド, 更新レートのコマンド)
        """
        # A342では設定済み
        assert self.rms_pp_interval

        # config取得時に検証済みだが一応チェック
        if self.rms_pp_interval < 1 or self.rms_pp_interval > 255:
            raise ValueError

        # 生データ出力時は使用しないためデフォルト値を返す
        if self.mode == Constants.MODE_RAW:
            return (0x0A, 0x07)

        if self.rms_pp_interval <= 1:
            update_rate = 4
        elif self.rms_pp_interval <= 3:
            update_rate = 5
        elif self.rms_pp_interval <= 6:
            update_rate = 6
        elif self.rms_pp_interval <= 13:
            update_rate = 7
        elif self.rms_pp_interval <= 27:
            update_rate = 8
        elif self.rms_pp_interval <= 54:
            update_rate = 9
        elif self.rms_pp_interval <= 109:
            update_rate = 10
        elif self.rms_pp_interval <= 218:
            update_rate = 11
        else:
            update_rate = 12

        return (self.rms_pp_interval, update_rate)

    def _get_record_per_sec(self) -> int:
        if self.mode == Constants.MODE_RAW:
            return self.sps

        assert self.rms_pp_interval
        # 出力間隔(秒)
        interval = (
            self.rms_pp_interval
            if self.physical == Constants.PHYSICAL_DISPLACEMENT
            else self.rms_pp_interval / 10
        )
        # サンプリングレートは出力間隔の逆数
        sps = 1 / interval

        return math.ceil(sps)

    def _get_record_per_file(self) -> int:
        if self.mode == Constants.MODE_RAW:
            return self.sps * self.file_rotate_min * Constants.SEC_IN_MINUTE

        assert self.rms_pp_interval

        # 出力間隔(秒)
        interval = (
            self.rms_pp_interval
            if self.physical == Constants.PHYSICAL_DISPLACEMENT
            else self.rms_pp_interval / 10
        )
        # サンプリングレートは出力間隔の逆数
        sps = 1 / interval

        return math.ceil((sps * self.file_rotate_min * Constants.SEC_IN_MINUTE))

    def _get_complement_data_params(self) -> Tuple[int, int]:
        """A342ReaderArgsでのデータ欠損補完に用いるパラメータを取得するメソッド

        Returns:
            Tuple[int, int]: 1番目:index0の時のcount値、2番目: カウントの増分
        """
        # 生データの時とRMS, P-Pで2bitカウンタの時にはコンストラクタで取得した値をそのまま使える
        if self.mode == Constants.MODE_RAW or self.baud == Constants.BAUD_RATE_460800:
            return (self.count_diff, self.count_diff)
        # RMS, P-Pで921.6kbpsの時はカウントの増分がカウントの最大値を超えうるため、仮想的な最大値を計算しておく
        else:
            assert self.rms_pp_interval
            count_diff = self.rms_pp_interval * 600
            count_start = (count_diff + 2) % self.count_max
            return (count_start, count_diff)
